"""
    QuaO Project post_processing_promise.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..callback.callback_url import CallbackUrl
from ..response.authentication import Authentication
from ...data.promise.promise import Promise


class PostProcessingPromise(Promise):
    def __init__(
        self, callback_url: CallbackUrl, authentication: Authentication, job_result
    ):
        super().__init__(callback_url, authentication)
        self.job_result = job_result



"""
    QuaO Project invocation_handler.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""

from ..component.backend.backend import Backend
from ..config.thread_config import circuit_running_pool
from ..data.request.invocation_request import InvocationRequest
from ..data.response.job_response import JobResponse
from ..enum.invocation_step import InvocationStep
from ..handler.handler import Handler
from ..util.response_utils import ResponseUtils


class InvocationHandler(Handler):
    def __init__(self, request_data: dict):
        super().__init__(request_data)

    def invoke(self, circuit_preparation_fn, post_processing_fn):
        """

        @param post_processing_fn: post processing function
        @param circuit_preparation_fn: circuit preparation function
        @return:
        """

        invocation_request = InvocationRequest(self.request_data)

        backend = Backend(invocation_request)

        circuit_running_pool.submit(backend.submit_job,
                                    circuit_preparation_fn,
                                    post_processing_fn)

        promise_response = JobResponse(authentication=backend.authentication,
                                       invocation_step=InvocationStep.PROMISE)

        return ResponseUtils.generate_response(promise_response)



"""
    QuaO Project response_utils.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from .job_status_mapping_utils import JobStatusMappingUtils
from .status_code_mapping_utils import StatusCodeMappingUtils
from ..data.response.job_response import JobResponse


class ResponseUtils:

    @staticmethod
    def generate_response(job_response: JobResponse) -> dict:
        if job_response:

            job_dict = {
                "providerJobId": job_response.provider_job_id,
                "jobStatus": job_response.job_status,
                "jobResult": job_response.job_result,
                "contentType": job_response.content_type.value,
                "histogram": job_response.job_histogram,
                "executionTime": job_response.execution_time
            }

            mapping_key = job_response.invocation_step \
                if JobStatusMappingUtils.resolve_job_status(job_response.job_status) is None \
                else job_response.job_status

            response = {
                "statusCode": StatusCodeMappingUtils.resolve_status_code(mapping_key).value,
                "body": job_dict,
                "userIdentity": job_response.user_identity,
                "userToken": job_response.user_token
            }

        else:
            response = {
                "statusCode": 500,
                "body": "Error in function code. Please contact the developer.",
                "userIdentity": job_response.user_identity,
                "userToken": job_response.user_token
            }

        print(response)

        return response



"""
    QuaO Project handler_factory.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from ..handler.invocation_handler import InvocationHandler
from ..handler.job_fetching_handler import JobFetchingHandler


class HandlerFactory:
    @staticmethod
    def create_handler(event):
        request_data = event.json()
        provider_job_id = request_data.get("provider_job_id")

        if provider_job_id is None:
            return InvocationHandler(request_data)
        return JobFetchingHandler(request_data)



"""
    QuaO Project quao_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
from qiskit import transpile
from qiskit import QiskitError

from ...data.device.circuit_running_option import CircuitRunningOption
from ...enum.status.job_status import JobStatus
from ...enum.processing_unit import ProcessingUnit
from ...enum.sdk import Sdk
from ..device.device import Device
from ..provider.provider import Provider
from ...config.logging_config import logger


class QuaoDevice(Device):

    def __init__(self, provider: Provider,
                 device_specification: str,
                 sdk: str,
                 processing_unit: ProcessingUnit):
        super().__init__(provider, device_specification)
        self.sdk = sdk
        self.processing_unit = processing_unit

    def _get_name(self) -> str:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return str(self.device.configuration().backend_name)

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return str(self.device.name)

        raise Exception("Sdk not supported!")

    def _create_job(self, circuit,  options: CircuitRunningOption):
        logger.debug('Create Quao job with {0} shots'.format(options.shots))
        if Sdk.QISKIT.value.__eq__(self.sdk):
            self.device.set_options(device=self.processing_unit.value, shots=options.shots)
            transpiled_circuit = transpile(circuits=circuit, backend=self.device)

            return self.device.run(transpiled_circuit)

        if Sdk.BRAKET.value.__eq__(self.sdk):
            import time

            start_time = time.time()

            job = self.device.run(task_specification=circuit, shots=options.shots)

            self.execution_time = time.time() - start_time

            return job

        raise Exception("Sdk not supported!")

    def _is_simulator(self) -> bool:
        return True

    def _produce_histogram_data(self, job_result) -> dict:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            try:
                histogram_data = job_result.get_counts()
            except QiskitError as qiskit_error:
                logger.debug("Can't produce histogram with error: {0}".format(str(qiskit_error)))
                histogram_data = None

            return histogram_data

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return dict(job_result.measurement_counts)

    def _get_provider_job_id(self, job) -> str:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return job.job_id()

        if Sdk.BRAKET.value.__eq__(self.sdk):
            return job.id

    def _get_job_status(self, job) -> str:
        if Sdk.QISKIT.value.__eq__(self.sdk):
            return job.status().name

        if Sdk.BRAKET.value.__eq__(self.sdk):
            job_state = job.state()
            if JobStatus.COMPLETED.value.__eq__(job_state):
                job_state = JobStatus.DONE.value
            return job_state

    def _calculate_execution_time(self, job_result):
        if Sdk.QISKIT.value.__eq__(self.sdk):
            if "_metadata" not in job_result:
                return None
    
            metadata = job_result["_metadata"]["metadata"]
    
            if (
                metadata is None
                or not bool(metadata)
                or "time_taken_execute" not in metadata
            ):
                return None
    
            self.execution_time = metadata["time_taken_execute"]
